package database

import (
	"chromagies/src/config"
	"chromagies/src/profiler"
	"database/sql"
	"fmt"
	"log"
	"strings"
	"time"

	_ "github.com/go-sql-driver/mysql"
)

const (
	profiler_DATABASE_CONNECT string = "Database Connect"
	profiler_DATABASE_QUERY   string = "Database Query"
)

const (
	database_ROUTINE_TYPE_PROCEDURE uint8 = 1
	database_ROUTINE_TYPE_FUNCTION  uint8 = 2
)

type databaseEnumIndexType string

const (
	database_ENUM_INDEX_PRIMARY databaseEnumIndexType = "PRIMARY"
	database_ENUM_INDEX_UNIQUE  databaseEnumIndexType = "UNIQUE"
)

type columsDefinition struct {
	Name       string
	Type       string
	Default    any
	IsNullable bool
	Extra      string
	Charset    any
	Collation  any
	Table      string
}

type indexDefinition struct {
	Type    databaseEnumIndexType
	Name    string
	Table   string
	Columns []string
}

type foreignKeyDefinition struct {
	Name             string
	Table            string
	ColumnName       string
	PointingToTable  string
	PointingToColumn string
	UpdateRule       string
	DeleteRule       string
}

type tableDefinition struct {
	Name    string
	Columns []columsDefinition
}

type databaseDefinition struct {
	Name        string
	Tables      []tableDefinition
	Indexes     []indexDefinition
	ForeignKeys []foreignKeyDefinition
}

var listTables = []tableDefinition{
	{
		Name: "users",
		Columns: []columsDefinition{
			{
				Name:       "id",
				Type:       "int(11)",
				Default:    nil,
				IsNullable: false,
				Extra:      "AUTO_INCREMENT",
			},
			{
				Name:       "name",
				Type:       "varchar(256)",
				Default:    nil,
				IsNullable: false,
				Extra:      "",
				Charset:    "utf8mb4",
				Collation:  "utf8mb4_unicode_ci",
			},
			{
				Name:       "password",
				Type:       "varchar(256)",
				Default:    nil,
				IsNullable: false,
				Extra:      "",
				Charset:    "utf8mb4",
				Collation:  "utf8mb4_unicode",
			},
			{
				Name:       "password-salt",
				Type:       "varchar(256)",
				Default:    nil,
				IsNullable: false,
				Extra:      "",
				Charset:    "utf8mb4",
				Collation:  "utf8mb4_unicode",
			},
			{
				Name:       "permission_level",
				Type:       "enum('ADMIN', 'MODERATOR', 'USER')",
				Default:    "USER",
				IsNullable: false,
				Extra:      "",
				Charset:    nil,
			},
			{
				Name:       "email",
				Type:       "varchar(256)",
				Default:    nil,
				IsNullable: true,
				Extra:      "",
				Charset:    "utf8mb4",
				Collation:  "utf8mb4_unicode",
			},
		},
	},
}

var dbPath string
var dbConfig config.Database
var isInit bool
var storedRoutines map[string]uint8
var db *sql.DB

func Init() {
	if isInit {
		return
	}

	dbConfig = config.GetConfig().Database
	dbPath = fmt.Sprintf("%s:%s@(%s:%d)/%s", dbConfig.User, dbConfig.Password, dbConfig.Host, dbConfig.Port, dbConfig.Database)

	profiler.Register(profiler_DATABASE_CONNECT, 200, "µs")
	profiler.Register(profiler_DATABASE_QUERY, 4000, "µs")

	isInit = true
	initDatabaseStructure()
}

func connect() {
	var err error
	log.Printf("Connecting to database %s at %s:%d as %s\n", dbConfig.Database, dbConfig.Host, dbConfig.Port, dbConfig.User)

	tStart := time.Now()
	for i := range dbConfig.RetriesOnError {
		db, err = sql.Open("mysql", dbPath)
		if err == nil {
			break
		}
		log.Printf("Error while connecting the database [%d/%d] : %s\n", i+1, dbConfig.RetriesOnError, err)
		if i+1 == dbConfig.RetriesOnError {
			break
		}
		time.Sleep((time.Duration)(dbConfig.TimeBetweenRetriesMs) * time.Millisecond)
	}

	if err != nil {
		log.Fatalln("Cannot connect to database !")
	}

	db.SetConnMaxLifetime(time.Minute * 1)
	db.SetMaxOpenConns(100)
	db.SetMaxIdleConns(100)

	ping()

	profiler.Add(profiler_DATABASE_CONNECT, time.Duration(time.Since(tStart).Microseconds()))
}

func ExecuteStoredRoutine(name string, args ...any) {
	if !isInit {
		return
	}
	if storedRoutines == nil {
		fetchStoredRoutines()
	}

	var exist bool
	var routineType uint8
	routineType, exist = storedRoutines[name]
	if !exist {
		return
	}

	switch routineType {
	case database_ROUTINE_TYPE_PROCEDURE:
		callStoredProcedure(name, args...)
	case database_ROUTINE_TYPE_FUNCTION:
		callStoredFunction(name, args...)
	}
}

func ping() {
	var err error
	for i := range dbConfig.RetriesOnError {
		err = db.Ping()
		if err == nil {
			break
		}
		log.Printf("Error while pinging the database [%d/%d] : %s\n", i+1, dbConfig.RetriesOnError, err)
		if i+1 == dbConfig.RetriesOnError {
			break
		}
		time.Sleep((time.Duration)(dbConfig.TimeBetweenRetriesMs) * time.Millisecond)
	}

	if err != nil {
		log.Fatalln("Cannot ping the database !")
	}
}

func executeQuery(query string, args ...any) *sql.Rows {
	if !isInit {
		return nil
	}
	if db == nil {
		connect()
	}

	var err error
	var rows *sql.Rows
	tStart := time.Now()
	for i := range dbConfig.RetriesOnError {
		rows, err = db.Query(query, args...)
		if err == nil {
			break
		}
		log.Printf("Error while querying the database [%d/%d] : %s\n", i+1, dbConfig.RetriesOnError, err)
		if i+1 == dbConfig.RetriesOnError {
			break
		}
		time.Sleep((time.Duration)(dbConfig.TimeBetweenRetriesMs) * time.Millisecond)
		ping()
	}
	profiler.Add(profiler_DATABASE_QUERY, time.Duration(time.Since(tStart).Microseconds()))

	if err != nil {
		log.Printf("Cannot query the database ! (Query : %s) with args ", query)
		log.Fatalln(args...)
	}

	return rows
}

func fetchStoredRoutines() {
	storedRoutines = nil
	var rows *sql.Rows = executeQuery("SELECT ROUTINE_NAME, ROUTINE_TYPE FROM information_schema.ROUTINES WHERE ROUTINE_SCHEMA=?;", dbConfig.Database)
	if rows != nil {
		storedRoutines = make(map[string]uint8, 0)
		defer rows.Close()
		for rows.Next() {
			var resName, resType string
			rows.Scan(&resName, &resType)
			switch resType {
			case "PROCEDURE":
				storedRoutines[resName] = database_ROUTINE_TYPE_PROCEDURE
			case "FUNCTION":
				storedRoutines[resName] = database_ROUTINE_TYPE_FUNCTION
			}
		}
	}
	fmt.Print("ROUTINES : ")
	fmt.Println(storedRoutines)
}

func callStoredProcedure(name string, args ...any) {
	var query string
	if len(args) == 0 {
		query = fmt.Sprintf("CALL %s()", name)
	} else if len(args) == 1 {
		query = fmt.Sprintf("CALL %s(?)", name)
	} else {
		query = strings.Repeat(", ?", len(args)-1)
		query = fmt.Sprintf("CALL %s(?%s)", name, query)
	}

	var rows *sql.Rows = executeQuery(query, args...)
	defer rows.Close()

	for rows.Next() {
		var res string
		rows.Scan(&res)
		log.Println(res)
	}
}

func callStoredFunction(name string, args ...any) {
	var query string
	if len(args) == 0 {
		query = fmt.Sprintf("SELECT %s()", name)
	} else if len(args) == 1 {
		query = fmt.Sprintf("SELECT %s(?)", name)
	} else {
		query = strings.Repeat(", ?", len(args)-1)
		query = fmt.Sprintf("SELECT %s(?%s)", name, query)
	}

	var rows *sql.Rows = executeQuery(query, args...)
	defer rows.Close()

	for rows.Next() {
		var res string
		rows.Scan(&res)
		log.Println(res)
	}
}

func fetchTables() []string {
	var rows *sql.Rows = executeQuery("SELECT TABLE_NAME FROM information_schema.TABLES WHERE TABLE_SCHEMA=?;", dbConfig.Database)
	var tables []string = make([]string, 0)
	if rows != nil {
		defer rows.Close()
		for rows.Next() {
			var resName string
			rows.Scan(&resName)
			tables = append(tables, resName)
		}
	}
	return tables
}

func fetchColumns(table string) []columsDefinition {
	var rows *sql.Rows = executeQuery("SELECT COLUMN_NAME, COLUMN_TYPE, COLUMN_DEFAULT, IS_NULLABLE, EXTRA, CHARACTER_SET_NAME, COLLATION_NAME FROM information_schema.COLUMNS WHERE TABLE_SCHEMA=? AND TABLE_NAME=?;", dbConfig.Database, table)
	var columns []columsDefinition = make([]columsDefinition, 0)
	if rows != nil {
		defer rows.Close()
		for rows.Next() {
			var resName, resType, resDefault, resNullable, resExtra, resCharset, resCollation string
			rows.Scan(&resName, &resType, &resDefault, &resNullable, &resExtra, &resCharset, &resCollation)
			columns = append(columns, columsDefinition{
				Name:       resName,
				Type:       resType,
				Default:    resDefault,
				IsNullable: resNullable == "YES",
				Extra:      resExtra,
				Charset:    resCharset,
				Collation:  resCollation,
				Table:      table,
			})
		}
	}
	return columns
}

func fetchIndexes() []indexDefinition {
	// SELECT * FROM STATISTICS as S INNER JOIN TABLE_CONSTRAINTS as TC ON TC.TABLE_SCHEMA = S.TABLE_SCHEMA AND TC.CONSTRAINT_NAME = S.INDEX_NAME AND TC.TABLE_NAME = S.TABLE_NAME WHERE S.TABLE_SCHEMA = "RPG";
	// SELECT S.COLUMN_NAME as COLUMN_NAME, S.INDEX_NAME AS INDEX_NAME, TC.CONSTRAINT_TYPE AS CONSTRAINT_TYPE FROM STATISTICS as S INNER JOIN TABLE_CONSTRAINTS as TC ON TC.TABLE_SCHEMA = S.TABLE_SCHEMA AND TC.CONSTRAINT_NAME = S.INDEX_NAME AND TC.TABLE_NAME = S.TABLE_NAME WHERE S.TABLE_SCHEMA = "RPG" AND S.TABLE_NAME = "folders" AND TC.CONSTRAINT_TYPE IN("UNIQUE", "PRIMARY KEY");

	// Add PRIMARY : ALTER TABLE <table> ADD CONSTRAINT PRIMARY KEY(<columns>);
	// Add UNIQUE : ALTER TABLE <table> ADD CONSTRAINT <constraint_name> UNIQUE(<columns>);

	var rows *sql.Rows = executeQuery("SELECT S.COLUMN_NAME as COLUMN_NAME, S.INDEX_NAME AS INDEX_NAME, TC.CONSTRAINT_TYPE AS CONSTRAINT_TYPE, S.TABLE_NAME AS TABLE_NAME FROM information_schema.STATISTICS as S INNER JOIN information_schema.TABLE_CONSTRAINTS as TC ON TC.TABLE_SCHEMA = S.TABLE_SCHEMA AND TC.CONSTRAINT_NAME = S.INDEX_NAME AND TC.TABLE_NAME = S.TABLE_NAME WHERE S.TABLE_SCHEMA = ? AND TC.CONSTRAINT_TYPE IN(\"UNIQUE\", \"PRIMARY KEY\");", dbConfig.Database)
	var indexes []indexDefinition = make([]indexDefinition, 0)
	if rows != nil {
		defer rows.Close()
		for rows.Next() {
			var resColumn, resIndex, resType, resTable string
			rows.Scan(&resColumn, &resIndex, &resType, &resTable)
			var index indexDefinition = indexDefinition{
				Table:   resTable,
				Name:    resIndex,
				Type:    database_ENUM_INDEX_PRIMARY,
				Columns: []string{resColumn},
			}
			if resType == "UNIQUE" {
				index.Type = database_ENUM_INDEX_UNIQUE
			}
			indexes = append(indexes, index)
		}
	}
	return indexes
}

func fetchForeignKeys() []foreignKeyDefinition {

	// ADD ALTER TABLE <table> ADD CONSTRAINT <constraint_name> FOREIGN KEY(<column>) REFERENCES <ref_table>(<ref_column>) ON DELETE <CASCADE / SET NULL> UPDATE <CASCADE / SET NULL>;

	// SELECT RC.CONSTRAINT_NAME as CONTRAINT_NAME, KCU.TABLE_NAME, KCU.COLUMN_NAME, KCU.REFERENCED_TABLE_NAME, KCU.REFERENCED_COLUMN_NAME, RC.UPDATE_RULE, RC.DELETE_RULE FROM REFERENTIAL_CONSTRAINTS AS RC INNER JOIN KEY_COLUMN_USAGE AS KCU ON KCU.CONSTRAINT_NAME = RC.CONSTRAINT_NAME WHERE KCU.TABLE_SCHEMA = "RPG";
	var rows *sql.Rows = executeQuery("SELECT RC.CONSTRAINT_NAME as CONTRAINT_NAME, KCU.TABLE_NAME, KCU.COLUMN_NAME, KCU.REFERENCED_TABLE_NAME, KCU.REFERENCED_COLUMN_NAME, RC.UPDATE_RULE, RC.DELETE_RULE FROM information_schema.REFERENTIAL_CONSTRAINTS AS RC INNER JOIN information_schema.KEY_COLUMN_USAGE AS KCU ON KCU.CONSTRAINT_NAME = RC.CONSTRAINT_NAME WHERE KCU.TABLE_SCHEMA = ?;", dbConfig.Database)
	var foreignKeys []foreignKeyDefinition = make([]foreignKeyDefinition, 0)
	if rows != nil {
		defer rows.Close()
		for rows.Next() {
			var resName, resTable, resColumn, resRefTable, resRefColumn, resUpdate, resDelete string
			rows.Scan(&resName, &resTable, &resColumn, &resRefTable, &resRefColumn, &resUpdate, &resDelete)
			var fk foreignKeyDefinition = foreignKeyDefinition{
				Name:             resName,
				Table:            resTable,
				ColumnName:       resColumn,
				PointingToTable:  resRefTable,
				PointingToColumn: resRefColumn,
				UpdateRule:       resUpdate,
				DeleteRule:       resDelete,
			}
			foreignKeys = append(foreignKeys, fk)
		}
	}
	return foreignKeys
}

func initDatabaseStructure() {
	fetchStoredRoutines()
	fetchIndexes()
	fetchForeignKeys()

	log.Println("Tables :")
	var tables []string = fetchTables()
	for i := range tables {
		var columns []columsDefinition = fetchColumns(tables[i])
		log.Printf(" - Table %s : \n", tables[i])
		for j := range columns {
			log.Printf("   - Column %s %s IsNullable:%t Default:%s Extra:%s Charset:%s/%s\n",
				columns[j].Name, columns[j].Type, columns[j].IsNullable, columns[j].Default, columns[j].Extra, columns[j].Charset, columns[j].Collation)
		}
	}

	log.Println("Indexes :")
	var indexes []indexDefinition = fetchIndexes()
	for i := range indexes {
		log.Printf(" - Index %s\n", indexes[i])
	}

	log.Println("Foreign keys :")
	var foreignKeys []foreignKeyDefinition = fetchForeignKeys()
	for i := range foreignKeys {
		log.Printf(" - Foreign key %s\n", foreignKeys[i])
	}
}
